

##  RISC-V 特权架构

RISC-V有三个特权模式，分别是M(machine)模式、S(supervisor)模式，以及U(user)模式。

我们内核设计的重心在S-mode。

## 机器模式(M)

是hart能执行的最高权限模式，对底层有着完全的使用权，是所有risc-v处理器都必须实现的权限模式。(RISC-V-Reader-Chinese-v2p1.pdf)https://github.com/plctlab/riscv-operating-system-mooc/blob/main/refs/RISC-V-Reader-Chinese-v2p1.pdf

M模式下必须处理一些中断和异常，在xv6（mit教学使用的操作系统）中，设计者简单实现了M模式，但将所有中断和异常都交给S模式去处理。同时，我们还了解到开源的OpenSBI已经实现了risc-v下的M模式，同时qemu模拟器的默认bios即是OpenSBI。在比赛的说明中，提到允许使用qemu模拟器的默认bios，因此我们决定将操作系统设计的重心放在S模式，M模式使用OpenSBI来实现。

ps：hart是指一个硬件线程，可以认为是一个独立的指令执行流，有点类似于cpu上核的概念，但目前cpu一个核可以执行多个控制流（超线程技术？）。																																

## OpenSBI

risc-v芯片加电后，会从地址0X80000000读取第一条指令，OpenSBI程序的起始地址就放在这里。OpenSBI完成各种初始化后，会跳转到地址0x80200000，这里就是我们操作系统内核的起始地址。

OpenSBI不仅起到了BootLoader的作用，他还实现了SBI接口规范(RISC-V Supervisor Binary Interface，是risc-v规定的S模式与M模式的标准接口）可以为内核(S mode)提供服务。 https://github.com/riscv/riscv-sbi-doc/blob/master/riscv-sbi.adoc#legacy-sbi-extension-extension-ids-0x00-through-0x0f



## 内核入口

linker.ld 文件：将内核镜像放在0X80200000。跳转到0X80200000后首先运行的是start.S汇编文件。

start.S为每个hart分配一个16KB的内核栈，之后跳转到Kernel_Init，该地址在kernel.c文件中定义，即函数void Kernel_Init(void)的地址，之后运行该函数。

之后我们应该在这个函数里完成各种初始化。

Makefile文件还不太会写，目前源文件也不太多，就先这样写了。在目录下输入指令： make run，即可进入qemu运行内核。

下一个目标是完成控制台的初始化，同时完成标准输入输出库（stdio.h），测试内核。

4.18 李荣



## 多核启动

今天完成多核启动和控制台、stdio。

OpenSBI完成hart的初始化后，在跳转到内核时，会通过寄存器a0将hartid传递给内核，我们的内核初始化函数可以利用这个参数完成多核启动。



## 控制台和stand io

SBI接口规范 https://github.com/riscv-non-isa/riscv-sbi-doc/blob/master/riscv-sbi.adoc

文件 sbi.h  sbi.c 实现了部分SBI接口

```c
int32_t sbi_console_putchar(int32_t ch){
    int32_t return_val;
    asm volatile (
        "mv a7, %[type]\n"
        "mv a0, %[ch]\n"
        "ecall\n"
        "mv %[return_val], a0"
        : [return_val] "=r" (return_val)
        : [type] "r" (SBI_CONSOLE_PUTCHAR), [ch] "r" (ch)
    );
    return return_val;
}


int32_t sbi_console_getchar(void){
    int32_t return_val;
    asm volatile (
        "mv a7, %[type]\n"
        "ecall\n"
        "mv %[return_val], a0"
        : [return_val] "=r" (return_val)
        : [type] "r" (SBI_CONSOLE_GETCHAR)
    );
    return return_val;
}
```

通过这两个接口我们可以实现控制台输出和输入。

之后我们在console.h console.c 文件中封装了这两个接口，接着在stdio.c中完成了字符串的输入和接收。当我将函数定义为putchar和getchar以及printf时，编译会报warnning，即使在gcc编译条件后加上 -nostdlib也不行，不得已就将上述和标准库重名的函数名加上_k（表示kernel）。

在stdio.c里实现了字符串输出，但在测试中，字符串的输入有问题，在qemu的命令行里无法输入任何字符，这个问题放在明天解决。

最后，多核启动好像还没怎么深入了解，这个也放在明天解决。

4.19 李荣



## 多核启动

[riscv-sbi-doc/riscv-sbi.adoc at master · riscv-non-isa/riscv-sbi-doc (github.com)](https://github.com/riscv-non-isa/riscv-sbi-doc/blob/master/riscv-sbi.adoc#legacy-sbi-extension-extension-ids-0x00-through-0x0f)

[(20条消息) 「RISC-V Arch」SBI 规范解读（下）_书香度年华的博客-CSDN博客](https://billy.blog.csdn.net/article/details/129203874?spm=1001.2101.3001.6650.7&utm_medium=distribute.pc_relevant.none-task-blog-2~default~BlogCommendFromBaidu~Rate-7-129203874-blog-129195552.235^v30^pc_relevant_default_base3&depth_1-utm_source=distribute.pc_relevant.none-task-blog-2~default~BlogCommendFromBaidu~Rate-7-129203874-blog-129195552.235^v30^pc_relevant_default_base3&utm_relevant_index=8)

openSBI多核启动，但只有主核会跳到内核入口地址，其他核进入wfi()状态（低功耗休眠）。

需要主核调用SBI接口来唤醒休眠的hart。

考虑到多线程需要加锁，这部分功能就放到以后实现。



## trap和计时器

在risc-v中，异常(exception)和中断(interrupt)统称为trap。其运作流程大概为CPU捕获异常或收到中断请求，通过stvec寄存器(Supervisor Trap Vector Base Address Register)找到trap处理程序的地址，跳转到trap处理函数，trap处理函数根据造成中断的原因调用不同的处理函数，完成trap处理后返回原函数。

事实上，系统调用就是通过trap机制实现的，用户程序使用ecall产生软中断，进入trap处理函数，trap处理函数再调用系统调用函数，完成后返回原地址。但考虑到系统调用实现需要完成页表的切换，而目前我们还并没有实现页表，因此系统调用的实现就放在以后完成。

trap过程的一些细节。

`sstatus`寄存器(Supervisor Status Register)记录这当前的状态。SIE是S-mode中断使能，置为0时会在s-mode中禁止所有中断。



![image-20230424000338989](./Record.assets/image-20230424000338989.png)



`scause`寄存器记录着中断发生的原因。

![image-20230424001731536](./Record.assets/image-20230424001731536.png)

`stvec`寄存器记录着中断向量表的地址，也就是trap的入口。在真正进入trap处理函数前，我们需要完成寄存器的保存，即上下文切换。

`sscratch`寄存器用来记录帧栈的位置，可以实现模式切换。、

`sie`寄存器指出那些中断必须处理，那些中断必须忽略。其中s-mode时间中断是第6bit。

![image-20230424011925750](./Record.assets/image-20230424011925750.png)

在实际实现里，我们在trapentry.S里实现了上下文切换，调用kerneltrap函数。在trap.c文件里实现了kerneltrap函数，其根据scause寄存器的值判断是中断还是异常，然后分别调用各自的处理函数。处理函数interrupt_handdle()以及exception_handler()分别根据scause的寄存器的值分别处理异常或中断。

在trap初始化函数中init_trap()，我们开启中断使能，将trap处理函数kerneltrap()的地址写入stvec寄存器中。

此外，我们还实现了时钟中断和时钟中断处理，这里参考了ucore的部分实现思路[滴答滴答(时钟中断) - ucore step by step (gitbook.io)](https://1790865014.gitbook.io/ucore-step-by-step/intro-2/3_clock)。



#### 遇到的问题

1、编译时报错，具体代码没记。原因是trapentry.S文件引用了riscv.h，而riscv.h文件中使用了一些汇编没有的而C有的关键字(typedef、#ifndef等)和函数定义(一些寄存器相关的函数)，解决方法是将这部分代码移到了新文件register.h中。

2、将sie寄存器和sstatus寄存器的SIE字段搞混了。sie寄存器用来记录那些中断需要处理，那些不需要。sstatus的SIE字段是表示s-mode的中断使能。前者用于时钟中断的初始化，后者用于开启或关闭s-mode全局中断。

4.23 李荣



## 页表机制 物理内存管理

memory.c memory.h

利用页表管理物理内存，物理内存的地址称为physical address（pa）。

将内存分为4k大小的页来管理，利用数据结构run来表示空闲页。

```C
struct run{
    struct run *next;
};
```

run恰好在一个页的开头，所以一个指向run的指针恰好等于run所在页的起始地址，然后run结构体里的next则可以指向下一个空闲页。

我们定义了一个数据结构来管理空闲页列表。

```c
struct{
	//上锁的信号量 mark
    // 空闲的物理块
    struct run *freelist;
}kmem;
```

其由空闲的页的列表和一个信号量组成。因为信号量互斥机制还未完成，这里暂时没实现信号量。

void memfree(uint64_t *pa*) 实现将pa指向的页清空并填满1防止悬空引用，之后将该页加入空闲页列表。

void* alloc_page()分配页，返回该页的地址。

void init_mem()完成内存页管理的初始化。它将内核的结束后的地址（end，定义在linker.ld）到物理内存最大地址之间的空间全部空间都划分为页，并加入空闲页列表。



## 虚拟内存

参考了xv6、ucore等内核，我们决定将核心态的内核虚拟地址直接映射为物理地址。其中0x8000 0000 到 0x8020 0000是OpenSBI的地址，0x8020 0000 之后的地址是我们可用的地址，0x8000 0000之前的地址映射到外设（这个好像是qemu自动完成的，我们就暂时没管这部分了，之后写sd card驱动时再回头看看）。

### Sv39

![image-20230430232440859](./Record.assets/image-20230430232440859.png)

risc-v有三种虚拟地址模式，sv32，sv39，sv48，前一种是32bit架构，后两者是64bit架构。目前比较流行的就是sv39。

Sv39模式下，虚拟地址是三级页表寻址。当前进程的三级页表框（page table）的地址存放在寄存器satp中。从satp取出三级页表框，根据虚拟地址（virtual address ，va）的VPN[2]，找到页表项（page table entry，pte），pte前十位是标志位，之后的24位指向下一级页表框。以此类推，最后一级的页表框的页表项包含着物理地址，这就是从va获取pa的过程。函数pte_t* get_last_pgtable()实现了这个过程，其中alloc标志着在寻找的过程中缺页时是否分配页面。

satp寄存器的值如下：

![image-20230430235052035](./Record.assets/image-20230430235052035.png)

### 虚拟内存初始化

bool map_pages()实现了在一个页表中将给定的va和pa建立映射。

void init_kernelvm() 初始化内核虚拟内存，建立va到pa映射。

void init_hart_vm()将三级页表地址写入当前进程的satp寄存器，并刷新缓存（TLB）。



### 页面置换和缺页异常

内存空间不够的时候，操作系统将不常用的页面置换到外存上。当进程需要使用的页面不在内存中时，会触发缺页异常，同时通过页面置换算法将外存的页调入内存中来。

这部分明天完成。	4.30 李荣

这部分等进程管理、文件系统写完后再写。    5.1 李荣



## 多核启动和自旋锁

### 多核启动 hart.c hart.h kernel.c

OpenSBI会同时启动多核，多个核心去抢做主启动核，主启动核一个人去做共享资源的初始化，其他核心等待主启动核把共享资源初始化完，再做各自的初始化工作，之后使用wfi()挂起，等待主启动核的核间中断信号(Inter-Processor Interrupt, IPI)。

接着主启动核来到内核的入口地址0x8020 0000，即我们代码kernel.c 的函数Kernel_Init()（但在gdb调试时发现void Kernel_Init(uint64_t *hartid*)并不是严格在0x8020 0000这个位置，而是在稍后几行）。OpenSBI会将该核的hartid放在a0寄存器上。

需要注意的是，主启动核的hartid并不一定是0，而是随机的一个核，哪个核启动的快就会被选为主启动核，所以不能使用hartid来判断是否是主启动核。因为没有注意到这点，下意识的以为主启动核hartid一定为0，写完代码后找了很久的bug（当时看xv6的代码主启动核id就是0，但这是因为xv6自己实现了machine模式，指定id为0的核为主启动核，而OpenSBI采用lottery算法，选择最快的核作为主启动核）。这里我们使用一个标准`boot_flag`来区别主启动核和从核。

主启动核进入内核入口后，完成一系列的初始化，将`boot_flag`置为`true`，调用函数`wake_other_hart(boot_hartid)`启动其他从核。

```c
void wake_other_hart(int64_t boot_hartid){
    uint64_t hart_mask;
    for(int hartid = 0; hartid < KERNEL_MAX_NUM; hartid++){
        if(hartid != boot_hartid){
            sbi_hart_start((uint64_t)hartid, KERNEL_ENTRY, 0);
            hart_mask = 1L << hartid;
            sbi_send_ipi(&hart_mask);
        }
    }
}
```

`wake_other_hart(boot_hartid)`使用了OpenSBI提供的接口`void sbi_hart_start()` `void sbi_send_ipi()`,设置其他核心的状态和运行参数，再发送核间中断信号把从核唤醒。这两个接口与请参见[OpenSBI官方文档](https://github.com/riscv-non-isa/riscv-sbi-doc/blob/master/riscv-sbi.adoc) (Function: HART start (FID #0),  Extension: Send IPI (EID #0x04))。

从核启动后，也会来到内核的入口地址0x8020 0000，此时`boot_flag =true`，完成核自身的初始化即可。

### 自旋锁 spinlock.c spinlock.h hart.h

完成多核启动后，进入多线程就必须完成锁，不然printf的输出完全是混乱的。

上锁需要实现核关中断，因此我将自旋锁和hart绑定，自旋锁的数据结构如下：

```C
// spinlock.h
struct spinlock{
    uint32_t locked;            // 0 unlocked, 1 locked;
    char *name;
    struct hart *hart;
};
```

同时，一个进程或者hart可能有多个不同的自旋锁在工作。比如说，一个hart里调用函数a进入临界区，使用了自旋锁关中断，在临界区里a函数又调用了函数b，函数b也要使用自旋锁进入另一个临界区，这个时候出现了嵌套式的关中断。如果spinlock只是简单的关中断的话，函数a关中断，进入函数b再一次关中断，b出临界区后开中断，但函数a还没出临界区，中断就已经被打开了。

我使用了函数`push_interrupt()`和`push_interrupt()`，同时修改了hart的数据结构来实现这种嵌套关中断。

```C
// hart.h
struct hart{
    struct Process *pro;
    int interrupt_depth;         // 嵌套关中断的深度
    bool pre_interrupt;          // 关中断前的中断使能状态
};
```

```C
// spinlock.c
void push_interrupt(){
    struct hart *hart = myhart();
    uint64_t pre_interrupt = s_intr_get();
    s_intr_disable();
    if(hart->interrupt_depth == 0){
        hart->pre_interrupt = pre_interrupt;
    }
    (hart->interrupt_depth)++;
}

void pop_interrupt(){
    struct hart *hart = myhart();
    if(s_intr_get()){
        panic("pop_interrpt 1");
    }
    if(hart->interrupt_depth < 1){
        panic("pop_interrpt 2");
    }
    hart->interrupt_depth -= 1;
    if(hart->interrupt_depth == 0 && hart->pre_interrupt == true){
        s_intr_enable();
    }
}
```

每一次关中断，`hart->interrupt_depth`+1，每一次开中断该值-1，当其值为0时再恢复之前的中断使能状态。`hart->pre_interrupt`记录了之前的中断使能状态。

这样，完整的自旋锁上锁解锁函数如下：

```C
void acquire_lock(struct spinlock *lock){
    
    push_interrupt();
    // 判断hart是否持有锁,如果该hart已经持有该锁，则报错
    if(is_holding_lock(lock)){
        panic("reacquire");
    }
    //test and set
    while(__sync_lock_test_and_set(&lock->locked, 1) != 0) ;
    __sync_synchronize();
    lock->hart = myhart();
}

void release_lock(struct spinlock *lock){
    //holding
    if(!is_holding_lock(lock)){
        panic("release");
    }
    lock->hart = 0;
    __sync_synchronize();
    __sync_lock_release(&lock->locked);
    pop_interrupt();
}
```



自旋锁如何获取hartid？

```C
//hart.c
struct hart harts[6];       //最多六个核心,每个核心可以运行一个进程  mark 到时候宏定义一下

int32_t myhartid(){
    int32_t id = r_tp();
    return id;
}

struct hart* myhart(){
    int32_t id = myhartid();
    return &harts[id];
}

/*
 * @brief 初始化hart,将hartid写入tp寄存器
 * @param hartid 硬件线程id，最大为 KERNEL_MAX_NUM
*/
inline void init_hartid(uint64_t hartid){
    asm volatile(
        "mv tp, %0" 
        :
        : "r" (hartid % KERNEL_MAX_NUM)
    );
    harts[hartid].interrupt_depth = 0;
}
```

在hart进入内核最开始的时候，会调用`init_hartid()`将OpenSBI传来的参数hartid（存放在`a0`寄存器）写入到`tp`寄存器中。考虑到这里的硬件没有啥超线程技术，一个cpu核心就是一个hart，只能同时运行一个进程，因此我们将hartid写入到`tp`寄存器中。

spinlock调用`myhart()`就可以得到hart，从而实现开关中断和hart绑定。

### 遇到的问题

代码完成后bug非常多，从输出结果来看`spinlock`貌似没有工作，printf的输出总是混在一起。通过在关键位置输出当前hart的中断使能状态发现，关中断函数`s_intr_disable()`因为马虎写错了，并没有实现关中断。

误以为主启动核的hartid一定为0。通过输出hartid发现主启动核其实是不确定的。在网上查阅资料学习到了OpenSBI的多核启动基本逻辑（[多核启动基本逻辑分析 ](https://wangzhou.github.io/多核启动基本逻辑分析/)）。

修改完后多核启动还是有问题。当设置核心数为3时，内核可以很好的运行。但当核心数为4、5、6的时候，启动是有问题的，有时会直接卡死在主启动核，有时会开始在其他从核，同时printf原本确定的、完成的输出也会变得混乱，看起来自旋锁还是有问题。因为每次结果不确定，大部分时候根本没有输出，只好用GDB来调试。GDB编译和连接QEMU的过程还挺麻烦的，这部分过程记录在后面。经过GDB调试，发现`tp`寄存器的值有问题，和OpenSBI传入的值hartid（一开始存放在`a0`寄存器中）不相等。最后找到是函数`init_hartid()`中将` "r" (hartid % KERNEL_MAX_NUM)`的取余符号%写成了与符号&，导致写入`tp`寄存器的值有问题，自旋锁的功能也跟着出问题。



## 使用GDB调试

### 安装riscv64-unknown-elf-gdb

参考[GDB 调试方法](http://rcore-os.cn/rCore-Tutorial-deploy/docs/pre-lab/gdb.html)（比赛提供的镜像是ubuntu 22.04.2 LTS，完全按照这个教程有点问题）。

比赛提供的docker镜像[Docker 搭建 C 语言操作系统开发环境 - ToolDiy (ry.rs)](https://tooldiy.ry.rs/one_click_deployment/docker_os_c/)

比赛提供的镜像里没有risc-v版本的gdb `riscv64-unknown-elf-gdb`，我们需要自己编译gdb。

```shell
# 安装依赖
sudo apt-get install libncurses5-dev python python-dev texinfo libreadline-dev
sudo apt-get install libgmp-dev
```

前往[清华大学开源软件镜像站](https://mirrors.tuna.tsinghua.edu.cn/gnu/gdb/?C=M&O=D)下载gdb源代码，我下的是[gdb-12.1.tar.gz](https://mirrors.tuna.tsinghua.edu.cn/gnu/gdb/gdb-12.1.tar.gz)。

```shell
# 下载gdb源码
wget https://mirrors.tuna.tsinghua.edu.cn/gnu/gdb/gdb-12.1.tar.gz
# 解压
tar -xzvf gdb-12.1.tar.gz
# 进入目录 
cd gdb-12.1
mkdir build
cd build
# 没加 --with-python和 --enable-tui，加了因为python版本的问题会报错
../configure --prefix=/usr/local  --target=riscv64-unknown-elf --with-libgmp-prefix="path to gmp"
# 编译 安装
make -j$(nproc)
sudo make install
```



### 使用GDB+QEMU调试

```makefile
# qemu			 内核镜像文件		qemu参数
${QEMU} -kernel $(KERNEL_FILE) ${QEMUFLAGS} -s -S
```

运行这条命令后qemu会提供一个端口(localhost:1234)等待GDB的连接。

打开另一个控制台进入目录，输入下面命令即可。

```shell
riscv64-unknown-elf-gdb kernel.elf
target remote localhost:1234
```



ps：如果用的是vscode，点击垃圾桶左边那个分页就非常方便。

![image-20230507171027673](./Record.assets/image-20230507171027673.png)



### 更方便的方法

下载gdb-multiarch即可，不用编译



## 进程和内核栈

每个进程都有自己的内核栈和用户栈。怎么理解linux内核栈？ - 朱涵俊的回答 - 知乎 https://www.zhihu.com/question/57013926/answer/151472914



idle进程[一文搞懂linux cpu idle - 知乎 (zhihu.com)](https://zhuanlan.zhihu.com/p/548268554)



参考linux的进程设计，我们为每个hart设置一个`idle`进程（pid=0），优先级最低，当等待队列中没有进程的时候该进程会被唤醒执行，进入低功耗的循环，等待被抢占。

每个进程都有自己的内核栈和用户栈，进程的内核栈大小定义为两页，内核栈在进程创建的时候由alloc_page()随机分配，同时在内核栈的下一页的valid位置为0，但不映射物理地址，形成一个guard page。

不单独设置中断栈，我们将中断直接放在进程的内核栈处理。

以上都是幻想，因为想偷懒，不想实现kmalloc
