#include "riscv.h"

.global kerneltrap
.global kernel_trap_entry
.align 4
// 内核中断入口
/* 
 * 仅用于内核中的进程发生的中断，并没有用到trapframe
 * 直接用sp指针分配一个256字节的栈存储寄存器
*/
kernel_trap_entry:
    //上下文切换
    //分配栈空间
    addi sp, sp, -256
    //存储寄存器 
    sd ra, 0(sp)
    sd sp, 8(sp)
    sd gp, 16(sp)
    sd tp, 24(sp)
    sd t0, 32(sp)
    sd t1, 40(sp)
    sd t2, 48(sp)
    sd s0, 56(sp)
    sd s1, 64(sp)
    sd a0, 72(sp)
    sd a1, 80(sp)
    sd a2, 88(sp)
    sd a3, 96(sp)
    sd a4, 104(sp)
    sd a5, 112(sp)
    sd a6, 120(sp)
    sd a7, 128(sp)
    sd s2, 136(sp)
    sd s3, 144(sp)
    sd s4, 152(sp)
    sd s5, 160(sp)
    sd s6, 168(sp)
    sd s7, 176(sp)
    sd s8, 184(sp)
    sd s9, 192(sp)
    sd s10, 200(sp)
    sd s11, 208(sp)
    sd t3, 216(sp)
    sd t4, 224(sp)
    sd t5, 232(sp)
    sd t6, 240(sp)

    call kerneltrap

    // 恢复寄存器
    ld ra, 0(sp)
    ld sp, 8(sp)
    ld gp, 16(sp)
    // ld tp, 24(sp) 可能会切换hart
    ld t0, 32(sp)
    ld t1, 40(sp)
    ld t2, 48(sp)
    ld s0, 56(sp)
    ld s1, 64(sp)
    ld a0, 72(sp)
    ld a1, 80(sp)
    ld a2, 88(sp)
    ld a3, 96(sp)
    ld a4, 104(sp)
    ld a5, 112(sp)
    ld a6, 120(sp)
    ld a7, 128(sp)
    ld s2, 136(sp)
    ld s3, 144(sp)
    ld s4, 152(sp)
    ld s5, 160(sp)
    ld s6, 168(sp)
    ld s7, 176(sp)
    ld s8, 184(sp)
    ld s9, 192(sp)
    ld s10, 200(sp)
    ld s11, 208(sp)
    ld t3, 216(sp)
    ld t4, 224(sp)
    ld t5, 232(sp)
    ld t6, 240(sp)

    addi sp, sp, 256

    sret

/*
 * 新建立的内核进程/线程，switch后进入的内核入口点
 * 恢复设置/保存的寄存器，进入内核态
 * 注意使用到了trapframe（保存在a0）
*/
.text
.globl forkrets
.align 4
forkrets:
    move sp, a0

    //ld s1, 32*REGBYTES(sp)
    ld s2, 24(sp)
    //csrw sstatus, s1
    csrw sepc, s2

    ld ra, 40(sp)
    ld gp, 56(sp)
    ld tp, 64(sp)
    ld t0, 72(sp)
    ld t1, 80(sp)
    ld t2, 88(sp)
    ld s0, 96(sp)
    ld s1, 104(sp)
    ld a0, 112(sp)
    ld a1, 120(sp)
    ld a2, 128(sp)
    ld a3, 136(sp)
    ld a4, 144(sp)
    ld a5, 152(sp)
    ld a6, 160(sp)
    ld a7, 168(sp)
    ld s2, 176(sp)
    ld s3, 184(sp)
    ld s4, 192(sp)
    ld s5, 200(sp)
    ld s6, 208(sp)
    ld s7, 216(sp)
    ld s8, 224(sp)
    ld s9, 232(sp)
    ld s10, 240(sp)
    ld s11, 248(sp)
    ld t3, 256(sp)
    ld t4, 264(sp)
    ld t5, 272(sp)
    ld t6, 280(sp)

    ld sp, 48(sp)
    
    sret

.global kernel_thread_entry
/* 
 *  kernel_thread_entry是一个小跳板，他将arg(s1 register)移到a0，然后跳转到s0，
 *  即s0所指的函数。执行完后，跳转到do_exit()
 *  应用在process.c kernel_thread()
*/
.align 4
kernel_thread_entry:        # void kernel_thread_entry(void)
	move a0, s1
	jr s0

	jal do_exit