#ifndef _PROCESS_
#define _PROCESS_
#include "Type_defs.h"
#include "spinlock.h"
#include "memory.h"
#include "vm.h"

struct trapframe{
    uint64_t kernelSatp;
    uint64_t kernelSp;
    uint64_t trapHandler;
    uint64_t epc;                       // 之前的pc
    uint64_t kernelHartId;
    uint64_t ra;
    uint64_t sp;
    uint64_t gp;
    uint64_t tp;
    uint64_t t0;
    uint64_t t1;
    uint64_t t2;
    uint64_t s0;
    uint64_t s1;
    uint64_t a0;
    uint64_t a1;
    uint64_t a2;
    uint64_t a3;
    uint64_t a4;
    uint64_t a5;
    uint64_t a6;
    uint64_t a7;
    uint64_t s2;
    uint64_t s3;
    uint64_t s4;
    uint64_t s5;
    uint64_t s6;
    uint64_t s7;
    uint64_t s8;
    uint64_t s9;
    uint64_t s10;
    uint64_t s11;
    uint64_t t3;
    uint64_t t4;
    uint64_t t5;
    uint64_t t6;
};

// Saved registers for kernel context switches.
// 上下文切换
struct context{
    uint64_t ra;
    uint64_t sp;

    // callee-saved
    uint64_t s0;
    uint64_t s1;
    uint64_t s2;
    uint64_t s3;
    uint64_t s4;
    uint64_t s5;
    uint64_t s6;
    uint64_t s7;
    uint64_t s8;
    uint64_t s9;
    uint64_t s10;
    uint64_t s11;
};


//进程控制块 进程的几种状态
enum process_state{
    UNUSED,         //进程控制块未使用
    SLEEPING,       //阻塞 或 PCB被分配了但还未初始化
    RUNNABLE,       //就绪
    RUNNING,        //运行
    ZOMBIE          //僵尸 等待父进程调用 wait 来回收子进程资源 
};

struct PCB{
    uint32_t pid;
    uint32_t priority;              //进程优先级
    struct trapframe trapframe;     //trap帧    用于陷入内核态
    struct context context;         //上下文
    struct spinlock lock;   
    enum process_state state;       //进程状态
    struct PCB *parent;             //父进程
    uint64_t runtime;               //运行时间
    uint64_t size_of_memory;        //内存空间大小
    pgtable_t pagetable;            //用户态下的进程页表，进程陷入内核态后异常处理等可能需要使用
    uint64_t kstack;                   //内核栈
    char* name;                     //debug用
};

typedef int32_t pid_t;
void init_proc();
void init_hart_idle(uint64_t hartid);
struct PCB* alloc_PCB(void);
void free_proc(struct PCB *pro);
struct PCB* myPCB();
bool init_PCB(struct PCB* p );
pid_t alloc_pid();
int32_t do_exit(int32_t error_code);
struct PCB* pid2PCB(pid_t pid);
void idle();
void proc_run(struct PCB *p);

void init_proc_test();
#endif