#ifndef _RISCV_
#define _RISCV_

#include "memlayout.h"

#define XLEN 64                     // 64位

#define SSTATUS_SPP (1L << 8)       // Previous mode, 1=Supervisor, 0=User
#define SSTATUS_SPIE (1L << 5)      // Supervisor Previous Interrupt Enable
#define SSTATUS_UPIE (1L << 4)      // User Previous Interrupt Enable
#define SSTATUS_SIE (1L << 1)       // Supervisor Interrupt Enable
#define SSTATUS_UIE (1L << 0)       // User Interrupt Enable
#define SIE_STIE    (1L << 5)       // SIE 寄存器 timer

#define MAXVA (1L << 38)            // 最大虚拟地址

#define PGROUNDUP(sz)  (((sz)+PGSIZE-1) & ~(PGSIZE-1))
#define PGROUNDDOWN(a) (((a)) & ~(PGSIZE-1))

#define PTE_V (1L << 0)             // valid位
#define PTE_R (1L << 1)             // 读
#define PTE_W (1L << 2)             // 写
#define PTE_X (1L << 3)             // 执行
#define PTE_U (1L << 4)             // user
#define PTE_G (1L << 6)             // global mapping 
/*The A bit indicates the virtual page has
been read, written, or fetched from since the last time the A bit was cleared. The D bit indicates
the virtual page has been written since the last time the D bit was cleared.
from riscv-privileged-20190608-1.pdf  p68
*/
#define PTE_A (1L << 6)             // access
#define PTE_D (1L << 7)             // dirty


#define PA2PTE(pa) ((((uint64_t)pa) >> 12) << 10)       //将pa转化为pte
#define PTE2PA(pte) (((pte) >> 10) << 12)               //将pte转为pa
#define PTE_FLAGS(pte) ((pte) & 0x3FF)                  // 获取pte的flags
#define PGVA_MASK   0x1FF   
#define PG_OFFSET   12                                  //page offset
#define PPN_OFFSET  9                                   // PPN 9bit

// RISC-V-Reader-Chinese-v2p1.pdf p108
// satp mode域 0：不使用  8：sv39  9：sv48
#define SATP_MODE   (8L<<60)         

#endif