#include "sbi.h"

static inline uint64_t sbi_call(uint64_t sbi_EID, uint64_t sbi_FID,uint64_t arg0, uint64_t arg1, uint64_t arg2){
    uint64_t retval;
    asm volatile (
        "mv a7, %[sbi_EID]\n"
        "mv a6, %[sbi_FID]\n"
        "mv a0, %[arg0]\n"
        "mv a1, %[arg1]\n"
        "mv a2, %[arg2]\n"
        "ecall\n"
        "mv %[ret_val], a0"
        : [ret_val] "=r" (retval)
        : [sbi_EID] "r" (sbi_EID),  [sbi_FID] "r" (sbi_FID), [arg0] "r" (arg0), [arg1] "r" (arg1), [arg2] "r" (arg2)
        : "memory"
    );
    return retval;
}

int64_t sbi_console_putchar(int64_t ch){
    return sbi_call(SBI_CONSOLE_PUTCHAR, 0, (uint64_t)ch, 0, 0);
}


int64_t sbi_console_getchar(void){
    return sbi_call(SBI_CONSOLE_GETCHAR, 0, 0, 0, 0);

}

int64_t sbi_set_timer(uint64_t stime_value){
    return sbi_call(SBI_SET_TIMER, 0, stime_value, 0, 0);
}

inline void sbi_hart_start(uint64_t hartid, uint64_t start_addr, uint64_t opaque){
    //https://github.com/riscv-non-isa/riscv-sbi-doc/blob/master/riscv-sbi.adoc
    //Table 25. HSM Function List
    //opaque参数会返回到a1寄存器，但这里就没实现返回值struct sbiret了
    sbi_call(SBI_HART_START_EID, SBI_HART_START_FID, hartid, start_addr, 0);
}

/* 谢谢你， chatgpt
// 正在为您搜索：OpenSBI、接口、sbi_send_ipi、hart_mask
// 在 OpenSBI 中，sbi_send_ipi 接口用于发送 IPI（Inter-Processor Interrupt）中断到指定的 hart。其中，hart_mask 参数用于指定目标 hart 的 ID，它是一个位掩码（bitmask）类型的参数，每个位表示一个 hart 是否为目标，例如：
// 如果 hart_mask 为 0x1，则意味着仅向 hart 0 发送中断；
// 如果 hart_mask 为 0x3，则意味着向 hart 0 和 hart 1 发送中断。
// 具体来说，第 i 位的值为 1 则表示目标中含有 ID 为 i-1 的 hart，反之则表示不包含。
// 需要注意的是，sbi_send_ipi 接口中传入的 hart_mask 必须是一个合法的位掩码。如果通过 sbi_hart_get_count 接口获取到的 hart 数量为 n，则最高位的位置为 n-1，因此 hart_mask 应该小于 2^(n-1)。
// 参考资料：https://github.com/riscv-non-isa/riscv-sbi-doc/blob/master/riscv-sbi.adoc
*/
inline void sbi_send_ipi(uint64_t* hart_mask){
    sbi_call(SBI_SEND_IPI, 0, (uint64_t)hart_mask, 0, 0);
}