#ifndef _SBI_
#define _SBI_
/*
* sbi接口规范 
* https://github.com/riscv-non-isa/riscv-sbi-doc/blob/master/riscv-sbi.adoc
*/

#include "Type_defs.h"

#define SBI_SET_TIMER               0
#define SBI_CONSOLE_PUTCHAR         1
#define SBI_CONSOLE_GETCHAR         2
#define SBI_CLEAR_IPI               3
#define SBI_SEND_IPI                4
#define SBI_REMOTE_FENCE_I          5
#define SBI_REMOTE_SFENCE_VMA       6
#define SBI_REMOTE_SFENCE_VMA_ASID  7
#define SBI_SHUTDOWN                8

#define SBI_HART_START_EID          0x48534DL
#define SBI_HART_START_FID          0

// //Function Name     SBIVersion  FID     EID
// //sbi_send_ipi      0.2         0       0x735049
// #define SBI_SEND_IPI_FID            0
// #define SBI_SEND_IPI_EID            0x735049L

struct sbiret{
    int64_t error;
    int64_t value;
};

//基础扩展   暂时不实现
// https://github.com/riscv-non-isa/riscv-sbi-doc/blob/master/riscv-sbi.adoc
struct sbiret sbi_get_spec_version(void);
struct sbiret sbi_get_impl_id(void);
struct sbiret sbi_get_impl_version(void);
struct sbiret sbi_probe_extension(long extension_id);
struct sbiret sbi_get_mvendorid(void);
struct sbiret sbi_get_marchid(void);
struct sbiret sbi_get_mimpid(void);

int64_t sbi_console_putchar(int64_t ch);
int64_t sbi_console_getchar(void);
int64_t sbi_set_timer(uint64_t stime_value);

void sbi_hart_start(uint64_t hartid,                             
                    uint64_t start_addr, 
                    uint64_t opaque);

void sbi_send_ipi(uint64_t *hart_mask);
#endif